/*
 * APM BIOS driver for Linux
 * by Stephen Rothwell
 *    sfr@pdact.pd.necisa.oz.au
 *
 * $Id: apm_bios.c,v 0.2 1994/06/02 06:56:09 sfr Exp $
 */

#include <asm/system.h>

#include <linux/timer.h>
#include <linux/apm_bios.h>

#define APM_CHECK_TIMEOUT	(HZ)

struct long_pointer {
	unsigned long	offset;
	unsigned short	segment;
};

extern void			time_init(void);

static void			apm_check_timeout(unsigned long);

struct apm_bios_info		apm_bios_info;

static asmlinkage struct long_pointer	apm_bios_entry;
static int				apm_enabled;
static int				clock_slowed;
static struct timer_list		apm_timer =
	{ NULL, NULL, APM_CHECK_TIMEOUT, 0, &apm_check_timeout };

static char				apm_driver_version[] = "0.2";

static void set_time(void)
{
	unsigned long	flags;

	save_flags(flags);
	cli();
	time_init();
	restore_flags(flags);
}

static unsigned short get_event(void)
{
	unsigned short	error;
	unsigned short	event;

	static int notified = 0;

	__asm__("lcall _apm_bios_entry\n\t"
		"setc %%al"
		: "=a" (error), "=b" (event)
		: "0" (0x530b));

	if ((error & 0xff) == 0)
		return event;

	if (error == 0x8001)
		return 0;

	if (notified++ == 0)
		printk("APM BIOS get_event error 0x%x\n", (error >> 8) & 0xff);
	return 0;
}

static int set_power_state(unsigned short state)
{
	unsigned short	error;

	__asm__("lcall _apm_bios_entry\n\t"
		"setc %%al"
		: "=a" (error)
		: "0" (0x5307), "1" (0x0001), "c" (state));
	if ((error & 0xff) != 0) {
		printk("APM BIOS set power state (%x) returned error %x\n",
			state, error >> 8);
		return 0;
	}
	return 1;
}

static void apm_check_timeout(unsigned long unused)
{
	unsigned short	event;

	while ((event = get_event()) != 0) {
		switch (event) {
		case APM_SYS_STANDBY:
		case APM_USER_STANDBY:
			printk("APM BIOS got standby notify\n");
			(void) set_power_state(0x001);
			break;
		case APM_SYS_SUSPEND:
		case APM_USER_SUSPEND:
		case APM_CRITICAL_SUSPEND:
			printk("APM BIOS got suspend notify\n");
			if (set_power_state(0x0002))
				set_time();
			break;
		case APM_NORMAL_RESUME:
		case APM_CRITICAL_RESUME:
			printk("APM BIOS got resume notify\n");
			set_time();
			break;
		case APM_UPDATE_TIME:
			printk("APM BIOS got update time\n");
			set_time();
			break;
		default:
			printk("APM BIOS got unknown event 0x%x\n", event);
			break;
		}
	}

	apm_timer.expires = APM_CHECK_TIMEOUT;
	add_timer(&apm_timer);
}

long apm_bios_init(long kmem_start)
{
	unsigned short	ax;
	unsigned short	bx;
	unsigned short	cx;
	unsigned short	dx;
	unsigned short	error;
	char *		power_stat;
	char *		bat_stat;

	clock_slowed = 0;
	apm_enabled = 0;

	if (apm_bios_info.version == 0)
		return kmem_start;

	printk("APM BIOS ver %c.%c flags 0x%x (driver version %s)\n",
		((apm_bios_info.version >> 8) & 0xff) + '0',
		(apm_bios_info.version & 0xff) + '0',
		apm_bios_info.flags,
		apm_driver_version);

	if ((apm_bios_info.flags & APM_32_BIT_SUPPORT) == 0)
		return kmem_start;

	/*
	 * Fix up segment lengths for a 1.0 BIOS
	 */
	if (apm_bios_info.version == 0x100) {
		apm_bios_info.cseg_len = 0;
		apm_bios_info.dseg_len = 0;
	}

	/*
	 * Fix for the Compaq Contura 3/25c which reports BIOS version 0.1
	 * but is reportedly a 1.1 BIOS.
	 */
	if (apm_bios_info.version == 0x001) {
		apm_bios_info.version = 0x101;
	}

	printk("    entry %x:%lx cseg16 %x dseg %x cseg len %x, dseg len %x\n",
		apm_bios_info.cseg, apm_bios_info.offset, apm_bios_info.cseg_16,
		apm_bios_info.dseg, apm_bios_info.cseg_len,
		apm_bios_info.dseg_len);

	apm_enabled = 1;

	set_base(gdt[APM_CS >> 3],
		0xc0000000 + ((unsigned long)apm_bios_info.cseg << 4));
	set_base(gdt[APM_CS_16 >> 3],
		0xc0000000 + ((unsigned long)apm_bios_info.cseg_16 << 4));
	if (apm_bios_info.cseg_len == 0) {
		set_limit(gdt[APM_CS >> 3], 64 * 1024);
		set_limit(gdt[APM_CS_16 >> 3], 64 * 1024);
	} else {
		set_limit(gdt[APM_CS >> 3], apm_bios_info.cseg_len);
		set_limit(gdt[APM_CS_16 >> 3], apm_bios_info.cseg_len);
	}
	set_base(gdt[APM_DS >> 3],
		0xc0000000 + ((unsigned long)apm_bios_info.dseg << 4));
	if (apm_bios_info.dseg_len == 0)
		set_limit(gdt[APM_DS >> 3], 64 * 1024);
	else
		set_limit(gdt[APM_DS >> 3], apm_bios_info.dseg_len);

	apm_bios_entry.offset = apm_bios_info.offset;
	apm_bios_entry.segment = APM_CS;

	if (apm_bios_info.version > 0x100) {
		__asm__("lcall _apm_bios_entry\n\t"
			"setc %%bl"
			: "=a" (ax), "=b" (error)
			: "0" (0x530e), "1" (0), "c" (0x0101));
		if ((error & 0xff) != 0)
			apm_bios_info.version = 0x100;
	}

	__asm__("lcall _apm_bios_entry"
		: "=b" (bx), "=c" (cx), "=d" (dx)
		: "a" (0x530a), "0" (1));

	switch ((bx >> 8) & 0xff) {
	case 0:	power_stat = "off line"; break;
	case 1:	power_stat = "on line"; break;
	case 2: power_stat = "on backup power"; break;
	default: power_stat = "unknown"; break;
	}

	switch (bx & 0xff) {
	case 0: bat_stat = "high"; break;
	case 1: bat_stat = "low"; break;
	case 2: bat_stat = "critical"; break;
	case 3: bat_stat = "charging"; break;
	default: bat_stat = "unknown"; break;
	}

	printk("    AC %s, battery %s", power_stat, bat_stat);
	if ((cx & 0xff) == 0xff)
		printk(", battery life unknown\n");
	else
		printk(", battery life %d%%\n", cx & 0xff);
	if (apm_bios_info.version > 0x100) {
		printk("    battery flag = 0x%x, ", (cx >> 8) & 0xff);
		if (dx == 0xffff)
			printk("battery life unknown\n");
		else
			printk("battery life %d %s\n", dx & 0x7fff,
				((dx & 0x8000) == 0) ? "seconds" : "minutes");
	}

	add_timer(&apm_timer);

	return kmem_start;
}

void apm_bios_idle(void)
{
	unsigned short	error;

	if (!apm_enabled) {
		__asm__("hlt");
		return;
	}

	__asm__("lcall _apm_bios_entry\n\t"
		"setc %%al"
		: "=a" (error)
		: "0" (0x5305));

	if (error & 0xff)
		__asm__("hlt");
	else
		clock_slowed = (apm_bios_info.flags & APM_IDLE_SLOWS_CLOCK) != 0;
}

void apm_bios_busy(void)
{
	if (!clock_slowed)
		return;

	__asm__("lcall _apm_bios_entry"
		: /* no results */
		: "a" (0x5306));

	clock_slowed = 0;
}

struct file_operations apm_bios_fops = {
	NULL,		/* apm_bios_lseek */
	NULL,		/* apm_bios_read */
	NULL,		/* apm_bios_write */
	NULL,		/* apm_bios_readdir */
	NULL,		/* apm_bios_select */
	NULL,		/* apm_bios_ioctl */	
	NULL,		/* apm_bios_mmap */
	NULL,		/* apm_bios_open */
	NULL,		/* apm_bios_release */
	NULL,		/* apm_bios_fsync */
	NULL		/* apm_bios_fasync */
};
